import pytest
import numpy as np
import pandas as pd
from nglui import parser


def test_layer_names(test_state):
    layer_names = parser.layer_names(test_state)
    assert len(layer_names) == 3
    assert 'synapses' in layer_names


def test_layers(test_state):
    img_layer = parser.image_layers(test_state)
    assert len(img_layer) == 1
    assert img_layer[0] == 'imagery'

    seg_layers = parser.segmentation_layers(test_state)
    assert len(seg_layers) == 1
    assert seg_layers[0] == 'segments'

    anno_layers = parser.annotation_layers(test_state)
    assert len(anno_layers) == 1
    assert anno_layers[0] == 'synapses'

    lyr = parser.get_layer(test_state, anno_layers[0])
    assert isinstance(lyr, dict)
    assert lyr['name'] == anno_layers[0]


def test_tag_dictionary(test_state):
    tags = parser.tag_dictionary(test_state, 'synapses')
    assert tags[1] == 'ChC'


def test_view_settings(test_state):
    view = parser.view_settings(test_state)
    assert view['perspectiveZoom'] == 1500


def test_annotation_parsing(test_state):
    points = parser.point_annotations(test_state, 'synapses')
    assert len(points) == 6
    assert points[2][2] == 1440

    points, desc, tags = parser.point_annotations(test_state, 'synapses',
                                                  description=True, tags=True)
    assert len(desc) == len(tags)
    assert desc[1] is None
    assert tags[4][0] == 1



from nglui import EasyViewer
from nglui.easyviewer.utils import default_neuroglancer_base
import pandas as pd
import numpy as np
from collections.abc import Collection
from IPython.display import HTML
from .utils import bucket_of_values

DEFAULT_VIEW_KWS = {
    "layout": "xy-3d",
    "zoom_image": 2,
    "show_slices": False,
    "zoom_3d": 2000,
}


class StateBuilder:
    """A class for schematic mapping data frames into neuroglancer states.
    Parameters
    ----------
    """

    def __init__(
        self,
        layers=[],
        base_state=None,
        url_prefix=None,
        state_server=None,
        resolution=[4, 4, 40],
        view_kws={},
        client=None,
    ):
        if client is not None:
            if state_server is None:
                state_server = client.state.state_service_endpoint
            if url_prefix is None:
                url_prefix = client.info.viewer_site()
        if url_prefix is None:
            url_prefix = default_neuroglancer_base

        self._base_state = base_state
        self._layers = layers
        self._resolution = resolution
        self._url_prefix = url_prefix
        self._state_server = state_server

        base_kws = DEFAULT_VIEW_KWS.copy()
        base_kws.update(view_kws)
        self._view_kws = base_kws

    def _reset_state(self, base_state=None, compatibility_mode=False):
        """
        Resets the neuroglancer state status to a default viewer.
        """
        if base_state is None:
            base_state = self._base_state
        self._temp_viewer = EasyViewer(compatibility_mode=compatibility_mode)
        self._temp_viewer.set_state(base_state)

    def initialize_state(self, base_state=None, compatibility_mode=False):
        """Generate a new Neuroglancer state with layers as needed for the schema.

        Parameters
        ----------
        base_state : str, optional
            Optional initial state to build on, described by its JSON. By default None.
        """
        self._reset_state(base_state, compatibility_mode=compatibility_mode)

        if self._state_server is not None:
            self._temp_viewer.set_state_server(self._state_server)

        if self._resolution is not None:
            self._temp_viewer.set_resolution(self._resolution)

        self._temp_viewer.set_view_options(**self._view_kws)

    def render_state(
        self,
        data=None,
        base_state=None,
        return_as="url",
        url_prefix=None,
        link_text="Neuroglancer Link",
        compatibility_mode=False,
    ):
        """Build a Neuroglancer state out of a DataFrame.

        Parameters
        ----------
        data : pandas.DataFrame, optional
            DataFrame to use as a point source. By default None, for which
            it will return only the base_state and any fixed values.
        base_state : str, optional
            Initial state to build on, expressed as Neuroglancer JSON. By default None
        return_as : ['url', 'viewer', 'html', 'json', 'dict', 'shared'], optional
            Choice of output types. Note that if a viewer is returned, the state is not reset.
                url : Returns the raw url describing the state
                viewer : Returns an EasyViewer object holding the state information
                html : Returns an HTML link to the url, useful for notebooks.
                json : Returns a JSON string describing the state.
                dict : Returns a dict version of the JSON state.
            By default 'url'
        url_prefix : str, optional
            Neuroglancer URL prefix to use. By default None, for which it will open with the
            class default.
        link_text : str, optional
            Text to use for the link when returning as html, by default 'Neuroglancer Link'
        Returns
        -------
        string or neuroglancer.Viewer
            A link to or viewer for a Neuroglancer state with layers, annotations, and selected objects determined by the data.
        """
        if base_state is None:
            base_state = self._base_state
        self.initialize_state(
            base_state=base_state, compatibility_mode=compatibility_mode
        )

        if url_prefix is None:
            url_prefix = self._url_prefix

        self._render_layers(
            data,
            compatibility_mode=compatibility_mode,
        )

        if return_as == "viewer":
            return self.viewer
        elif return_as == "url":
            url = self._temp_viewer.as_url(prefix=url_prefix)
            self.initialize_state()
            return url
        elif return_as == "html":
            out = self._temp_viewer.as_url(
                prefix=url_prefix, as_html=True, link_text=link_text
            )
            out = HTML(out)
            self.initialize_state()
            return out
        elif return_as == "dict":
            out = self._temp_viewer.state.to_json()
            self.initialize_state()
            return out
        elif return_as == "json":
            from json import dumps

            out = self._temp_viewer.state.to_json()
            self.initialize_state()
            return dumps(out)
        else:
            raise ValueError("No appropriate return type selected")

    def _render_layers(self, data, compatibility_mode):
        for layer in self._layers:
            layer._render_layer(
                self._temp_viewer,
                data,
                compatibility_mode=compatibility_mode,
                viewer_resolution=self._resolution,
            )

    @property
    def viewer(self):
        return self._temp_viewer


class ChainedStateBuilder:
    def __init__(self, statebuilders):
        """Builds a collection of states that sequentially add annotations based on a sequence of dataframes.

        Parameters
        ----------
        statebuilders : list
            List of DataStateBuilders, in same order as dataframes will be passed
        """
        self._statebuilders = statebuilders
        if len(self._statebuilders) == 0:
            raise ValueError("Must have at least one statebuilder")

    def render_state(
        self,
        data_list=None,
        base_state=None,
        return_as="url",
        url_prefix=default_neuroglancer_base,
        link_text="Neuroglancer Link",
        compatibility_mode=False,
    ):
        """Generate a single neuroglancer state by addatively applying an ordered collection of
        dataframes to an collection of StateBuilder renders.
        Parameters
            data_list : Collection of DataFrame. The order must match the order of StateBuilders
                        contained in the class on creation.
            base_state : JSON neuroglancer state (optional, default is None).
                         Used as a base state for adding everything else to.
            return_as: ['url', 'viewer', 'html', 'json']. optional, default='url'.
                       Sets how the state is returned. Note that if a viewer is returned,
                       the state is not reset to default.
            url_prefix: string, optional (default is https://neuromancer-seung-import.appspot.com).
                        Overrides the default neuroglancer url for url generation.
        """
        if data_list is None:
            data_list = len(self._statebuilders) * [None]

        if len(data_list) != len(self._statebuilders):
            raise ValueError("Must have as many dataframes as statebuilders")

        temp_state = base_state
        for builder, data in zip(self._statebuilders[:-1], data_list[:-1]):
            temp_state = builder.render_state(
                data=data,
                base_state=temp_state,
                return_as="dict",
                compatibility_mode=compatibility_mode,
            )
        last_builder = self._statebuilders[-1]
        return last_builder.render_state(
            data=data_list[-1],
            base_state=temp_state,
            return_as=return_as,
            url_prefix=url_prefix,
            link_text=link_text,
            compatibility_mode=compatibility_mode,
        )



from setuptools import setup
import re
import os
import codecs
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open('requirements.txt', 'r') as f:
    required = f.read().splitlines()

with open('test_requirements.txt', 'r') as f:
    test_required = f.read().splitlines()

setup(
    name='nglui',
    version=find_version('src', 'nglui', '__init__.py'),
    description='Framework for data-driven generation of neuroglancer states.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Derrick Brittain, Casey Schneider-Mizell, Forrest Collman',
    author_email='caseysm@gmail.com',
    url="https://github.com/seung-lab/NeuroglancerAnnotationUI",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    setup_requires=['pytest-runner'],
    tests_require=test_required,
    include_package_data=True,
    install_requires=[required],  # external packages as dependencies
)



import pytest
from nglui import annotation
import numpy as np


def test_viewer_layers(viewer, img_path, seg_path_precomputed, seg_path_graphene):
    viewer.add_image_layer("img", img_path)
    assert "img" in viewer.layer_names

    viewer.add_segmentation_layer("seg_pre", seg_path_precomputed)
    assert viewer.state.layers["seg_pre"].type == "segmentation"

    viewer.add_segmentation_layer("seg_graph", seg_path_graphene)
    assert viewer.state.layers["seg_graph"].type == "segmentation_with_graph"

    id_val = 1000
    viewer.add_selected_objects("seg_graph", [id_val])
    assert id_val in viewer.state.layers["seg_graph"].segments


def test_pose_and_navigation(viewer):
    viewer.set_resolution([8, 8, 40])
    assert np.array_equal(viewer.state.voxel_size, [8, 8, 40])

    new_pos = [3926, 3528, 4070]
    viewer.set_view_options(position=new_pos, zoom_image=4)
    assert all(
        new_pos[ii] == int(val)
        for ii, val in enumerate(viewer.state.position.voxel_coordinates)
    )


def test_annotations(viewer, anno_layer_name):
    anno_ln = anno_layer_name
    viewer.add_annotation_layer(layer_name=anno_ln, color="#00bb33")
    viewer.set_annotation_layer_color(anno_ln, color="#aabbcc")
    assert viewer.state.layers[anno_ln].annotationColor == "#aabbcc"

    pt_anno = annotation.point_annotation([1, 2, 3])
    line_anno = annotation.line_annotation([1, 2, 3], [2, 3, 4])
    sphere_anno = annotation.sphere_annotation([1, 2, 3], 4, 0.1)
    bb_anno = annotation.bounding_box_annotation([1, 2, 3], [2, 3, 4])
    viewer.add_annotations(anno_ln, [pt_anno, line_anno, sphere_anno, bb_anno])
    assert len(viewer.state.layers[anno_ln].annotations) == 4

    viewer.update_description({anno_ln: [pt_anno.id]}, "test_description")
    assert viewer.state.layers[anno_ln].annotations[0].description == "test_description"

    viewer.remove_annotations(anno_ln, [pt_anno.id])
    assert len(viewer.state.layers[anno_ln].annotations) == 3

    viewer.clear_annotation_layers([anno_ln])
    assert len(viewer.state.layers[anno_ln].annotations) == 0


def test_grouped_annotations(viewer, anno_layer_name):
    pt_anno_1 = annotation.point_annotation([1, 2, 3])
    pt_anno_2 = annotation.point_annotation([4, 5, 6])
    grp = annotation.group_annotations([pt_anno_1, pt_anno_2], return_all=False)
    assert grp.id == pt_anno_1.parent_id


def test_annotation_tags(viewer, anno_layer_name):
    anno_ln = anno_layer_name
    viewer.add_annotation_layer(layer_name=anno_ln)
    tags = ["tagA", "tagB"]
    tag_dict = {str(ii + 1): tag for ii, tag in zip(range(len(tags)), tags)}
    anno_A = annotation.point_annotation([1, 2, 3], tag_ids=[2])
    viewer.add_annotation_tags(anno_ln, tags)
    viewer.add_annotations(anno_ln, [anno_A])
    anno_a_ids = viewer.state.layers[anno_ln].annotations[0]._json_data["tagIds"]
    assert tag_dict[anno_a_ids[0]] == tags[1]
